@@ -70,18 +70,25 @@ impl<'source> Compiler<'source> {
     }
 
     /// Starts a for loop
-    pub fn start_for_loop(&mut self, target_name: &'source str) {
-        self.add(Instruction::PushLoop(target_name));
+    pub fn start_for_loop(&mut self, with_loop_var: bool) {
+        self.add(Instruction::PushLoop(with_loop_var));
         let iter_instr = self.add(Instruction::Iterate(!0));
         self.pending_block.push(PendingBlock::Loop(iter_instr));
     }
 
     /// Ends the open for loop
-    pub fn end_for_loop(&mut self) {
+    pub fn end_for_loop(&mut self, push_did_iterate: bool) {
         match self.pending_block.pop() {
             Some(PendingBlock::Loop(iter_instr)) => {
                 self.add(Instruction::Jump(iter_instr));
-                let loop_end = self.add(Instruction::PopFrame);
+                let loop_end = self.next_instruction();
+                if push_did_iterate {
+                    self.add(Instruction::Lookup("loop"));
+                    self.add(Instruction::GetAttr("index0"));
+                    self.add(Instruction::LoadConst(Value::from(0)));
+                    self.add(Instruction::Eq);
+                };
+                self.add(Instruction::PopFrame);
                 if let Some(Instruction::Iterate(ref mut jump_target)) =
                     self.instructions.get_mut(iter_instr)
                 {
@@ -179,12 +186,40 @@ impl<'source> Compiler<'source> {
             }
             ast::Stmt::ForLoop(for_loop) => {
                 self.set_location_from_span(for_loop.span());
-                self.compile_expr(&for_loop.iter)?;
-                self.start_for_loop(for_loop.target);
+
+                if let Some(ref filter_expr) = for_loop.filter_expr {
+                    // filter expressions work like a nested for loop without
+                    // the special loop variable that append into a new list
+                    // just outside of the loop.
+                    self.add(Instruction::BuildList(0));
+                    self.compile_expr(&for_loop.iter)?;
+                    self.start_for_loop(false);
+                    self.add(Instruction::DupTop);
+                    self.compile_assignment(&for_loop.target)?;
+                    self.compile_expr(filter_expr)?;
+                    self.start_if();
+                    self.add(Instruction::ListAppend);
+                    self.start_else();
+                    self.add(Instruction::DiscardTop);
+                    self.end_if();
+                    self.end_for_loop(false);
+                } else {
+                    self.compile_expr(&for_loop.iter)?;
+                }
+
+                self.start_for_loop(true);
+                self.compile_assignment(&for_loop.target)?;
                 for node in &for_loop.body {
                     self.compile_stmt(node)?;
                 }
-                self.end_for_loop();
+                self.end_for_loop(!for_loop.else_body.is_empty());
+                if !for_loop.else_body.is_empty() {
+                    self.start_if();
+                    for node in &for_loop.else_body {
+                        self.compile_stmt(node)?;
+                    }
+                    self.end_if();
+                }
             }
             ast::Stmt::IfCond(if_cond) => {
                 self.set_location_from_span(if_cond.span());
@@ -251,6 +286,25 @@ impl<'source> Compiler<'source> {
         Ok(())
     }
 
+    /// Compiles an assignment expression.
+    pub fn compile_assignment(&mut self, expr: &ast::Expr<'source>) -> Result<(), Error> {
+        match expr {
+            ast::Expr::Var(var) => {
+                self.set_location_from_span(var.span());
+                self.add(Instruction::StoreLocal(var.id));
+            }
+            ast::Expr::List(list) => {
+                self.set_location_from_span(list.span());
+                self.add(Instruction::UnpackList(list.items.len()));
+                for expr in &list.items {
+                    self.compile_assignment(expr)?;
+                }
+            }
+            _ => panic!("bad assignment target"),
+        }
+        Ok(())
+    }
+
     /// Compiles an expression.
     pub fn compile_expr(&mut self, expr: &ast::Expr<'source>) -> Result<(), Error> {
         match expr {
@@ -295,12 +349,26 @@ impl<'source> Compiler<'source> {
                     ast::BinOpKind::Rem => Instruction::Rem,
                     ast::BinOpKind::Pow => Instruction::Pow,
                     ast::BinOpKind::Concat => Instruction::StringConcat,
+                    ast::BinOpKind::In => Instruction::In,
                 };
                 self.set_location_from_span(c.span());
                 self.compile_expr(&c.left)?;
                 self.compile_expr(&c.right)?;
                 self.add(instr);
             }
+            ast::Expr::IfExpr(i) => {
+                self.set_location_from_span(i.span());
+                self.compile_expr(&i.test_expr)?;
+                self.start_if();
+                self.compile_expr(&i.true_expr)?;
+                self.start_else();
+                if let Some(ref false_expr) = i.false_expr {
+                    self.compile_expr(false_expr)?;
+                } else {
+                    self.add(Instruction::LoadConst(Value::UNDEFINED));
+                }
+                self.end_if();
+            }
             ast::Expr::Filter(f) => {
                 self.set_location_from_span(f.span());
                 self.compile_expr(&f.expr)?;
