@@ -27,8 +27,7 @@ pub enum ErrorKind {
     ImpossibleOperation,
     SyntaxError,
     TemplateNotFound,
-    InvalidFilterArguments,
-    InvalidTestArguments,
+    InvalidArguments,
     UnknownFilter,
     UnknownTest,
     BadEscape,
@@ -44,8 +43,7 @@ impl ErrorKind {
             ErrorKind::ImpossibleOperation => "impossible operation",
             ErrorKind::SyntaxError => "syntax error",
             ErrorKind::TemplateNotFound => "template not found",
-            ErrorKind::InvalidFilterArguments => "invalid filter arguments",
-            ErrorKind::InvalidTestArguments => "invalid test arguments",
+            ErrorKind::InvalidArguments => "invalid arguments",
             ErrorKind::UnknownFilter => "unknown filter",
             ErrorKind::UnknownTest => "unknown test",
             ErrorKind::BadEscape => "bad string escape",
