@@ -8,6 +8,9 @@ pub enum Instruction<'source> {
     /// Emits raw source
     EmitRaw(&'source str),
 
+    /// Stores a variable (only possible in for loops)
+    StoreLocal(&'source str),
+
     /// Load a variable,
     Lookup(&'source str),
 
@@ -26,6 +29,12 @@ pub enum Instruction<'source> {
     /// Builds a list of the last n pairs on the stack.
     BuildList(usize),
 
+    /// Unpacks a list into N stack items.
+    UnpackList(usize),
+
+    /// Appends to the list.
+    ListAppend,
+
     /// Add the top two values
     Add,
 
@@ -71,6 +80,9 @@ pub enum Instruction<'source> {
     /// String concatenation operator
     StringConcat,
 
+    /// Performs a containment check
+    In,
+
     /// Apply a filter.
     ApplyFilter(&'source str),
 
@@ -82,9 +94,9 @@ pub enum Instruction<'source> {
 
     /// Starts a loop
     ///
-    /// The argument is the variable name of the loop
-    /// variable.
-    PushLoop(&'source str),
+    /// The arugment is an indicator if the `loop` variable should
+    /// be available.
+    PushLoop(bool),
 
     /// Pushes a value as context layer.
     PushContext,
@@ -134,6 +146,12 @@ pub enum Instruction<'source> {
     /// Calls an object
     CallObject,
 
+    /// Duplicates the top item
+    DupTop,
+
+    /// Discards the top item
+    DiscardTop,
+
     /// A nop
     #[allow(unused)]
     Nop,
@@ -143,12 +161,15 @@ impl<'source> fmt::Debug for Instruction<'source> {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         match *self {
             Instruction::EmitRaw(s) => write!(f, "EMIT_RAW (string {:?})", s),
+            Instruction::StoreLocal(n) => write!(f, "STORE_LOCAL (var {:?})", n),
             Instruction::Lookup(n) => write!(f, "LOOKUP (var {:?})", n),
             Instruction::GetAttr(n) => write!(f, "GETATTR (key {:?})", n),
             Instruction::GetItem => write!(f, "GETITEM"),
             Instruction::LoadConst(ref v) => write!(f, "LOAD_CONST (value {:?})", v),
             Instruction::BuildMap(n) => write!(f, "BUILD_MAP ({:?} pairs)", n),
             Instruction::BuildList(n) => write!(f, "BUILD_LIST ({:?} items)", n),
+            Instruction::UnpackList(n) => write!(f, "UNPACK_LIST ({:?} items)", n),
+            Instruction::ListAppend => write!(f, "LIST_APPEND"),
             Instruction::Add => write!(f, "ADD"),
             Instruction::Sub => write!(f, "SUB"),
             Instruction::Mul => write!(f, "MUL"),
@@ -164,6 +185,7 @@ impl<'source> fmt::Debug for Instruction<'source> {
             Instruction::Lte => write!(f, "LTE"),
             Instruction::Not => write!(f, "NOT"),
             Instruction::StringConcat => write!(f, "STRING_CONCAT"),
+            Instruction::In => write!(f, "IN"),
             Instruction::ApplyFilter(n) => {
                 write!(f, "APPLY_FILTER (name {:?})", n)
             }
@@ -171,7 +193,7 @@ impl<'source> fmt::Debug for Instruction<'source> {
                 write!(f, "PERFORM_TEST (name {:?})", n)
             }
             Instruction::Emit => write!(f, "EMIT"),
-            Instruction::PushLoop(t) => write!(f, "PUSH_LOOP (assign to {:?})", t),
+            Instruction::PushLoop(v) => write!(f, "PUSH_LOOP (loop var: {:?})", v),
             Instruction::PushContext => write!(f, "PUSH_CONTEXT"),
             Instruction::Iterate(t) => write!(f, "ITERATE (exit to {:>05x})", t),
             Instruction::PopFrame => write!(f, "POP_FRAME"),
@@ -187,6 +209,8 @@ impl<'source> fmt::Debug for Instruction<'source> {
             Instruction::CallFunction(n) => write!(f, "CALL_FUNCTION (name {:?})", n),
             Instruction::CallMethod(n) => write!(f, "CALL_METHOD (name {:?})", n),
             Instruction::CallObject => write!(f, "CALL_OBJECT"),
+            Instruction::DupTop => write!(f, "DUP_TOP"),
+            Instruction::DiscardTop => write!(f, "DISCARD_TOP"),
             Instruction::Nop => write!(f, "NOP"),
         }
     }
