@@ -163,6 +163,35 @@ impl<'a> Parser<'a> {
         }
     }
 
+    fn parse_ifexpr(&mut self) -> Result<ast::Expr<'a>, Error> {
+        let mut span = self.stream.current_span();
+        let mut expr = self.parse_or()?;
+        loop {
+            if matches!(self.stream.current()?, Some((Token::Ident("if"), _))) {
+                self.stream.next()?;
+                let expr2 = self.parse_or()?;
+                let expr3 = if matches!(self.stream.current()?, Some((Token::Ident("else"), _))) {
+                    self.stream.next()?;
+                    Some(self.parse_ifexpr()?)
+                } else {
+                    None
+                };
+                expr = ast::Expr::IfExpr(Spanned::new(
+                    ast::IfExpr {
+                        test_expr: expr2,
+                        true_expr: expr,
+                        false_expr: expr3,
+                    },
+                    self.stream.expand_span(span),
+                ));
+                span = self.stream.current_span();
+            } else {
+                break;
+            }
+        }
+        Ok(expr)
+    }
+
     binop!(parse_or, parse_and, {
         Some((Token::Ident("or"), _)) => ast::BinOpKind::ScOr,
     });
@@ -177,6 +206,7 @@ impl<'a> Parser<'a> {
         let mut span = self.stream.current_span();
         let mut expr = self.parse_math1()?;
         loop {
+            let mut negated = false;
             let op = match self.stream.current()? {
                 Some((Token::Eq, _)) => ast::BinOpKind::Eq,
                 Some((Token::Ne, _)) => ast::BinOpKind::Ne,
@@ -184,9 +214,18 @@ impl<'a> Parser<'a> {
                 Some((Token::Lte, _)) => ast::BinOpKind::Lte,
                 Some((Token::Gt, _)) => ast::BinOpKind::Gt,
                 Some((Token::Gte, _)) => ast::BinOpKind::Gte,
+                Some((Token::Ident("in"), _)) => ast::BinOpKind::In,
+                Some((Token::Ident("not"), _)) => {
+                    self.stream.next()?;
+                    expect_token!(self, Token::Ident("in"), "in")?;
+                    negated = true;
+                    ast::BinOpKind::In
+                }
                 _ => break,
             };
-            self.stream.next()?;
+            if !negated {
+                self.stream.next()?;
+            }
             expr = ast::Expr::BinOp(Spanned::new(
                 ast::BinOp {
                     op,
@@ -195,6 +234,15 @@ impl<'a> Parser<'a> {
                 },
                 self.stream.expand_span(span),
             ));
+            if negated {
+                expr = ast::Expr::UnaryOp(Spanned::new(
+                    ast::UnaryOp {
+                        op: ast::UnaryOpKind::Not,
+                        expr,
+                    },
+                    self.stream.expand_span(span),
+                ));
+            }
             span = self.stream.current_span();
         }
         Ok(expr)
@@ -283,6 +331,11 @@ impl<'a> Parser<'a> {
                 }
                 Some((Token::Ident("is"), _)) => {
                     self.stream.next()?;
+                    let mut negated = false;
+                    if matches!(self.stream.current()?, Some((Token::Ident("not"), _))) {
+                        self.stream.next()?;
+                        negated = true;
+                    }
                     let (name, span) =
                         expect_token!(self, Token::Ident(name) => name, "identifier")?;
                     let args = if matches!(self.stream.current()?, Some((Token::ParenOpen, _))) {
@@ -294,6 +347,15 @@ impl<'a> Parser<'a> {
                         ast::Test { name, expr, args },
                         self.stream.expand_span(span),
                     ));
+                    if negated {
+                        expr = ast::Expr::UnaryOp(Spanned::new(
+                            ast::UnaryOp {
+                                op: ast::UnaryOpKind::Not,
+                                expr,
+                            },
+                            self.stream.expand_span(span),
+                        ));
+                    }
                 }
                 _ => break,
             }
@@ -339,7 +401,35 @@ impl<'a> Parser<'a> {
             Token::Int(val) => Ok(const_val!(val)),
             Token::Float(val) => Ok(const_val!(val)),
             Token::ParenOpen => {
-                let expr = self.parse_expr()?;
+                // MiniJinja does not really have tuples, but it treats the tuple
+                // syntax the same as lists.
+                if matches!(self.stream.current()?, Some((Token::ParenClose, _))) {
+                    self.stream.next()?;
+                    return Ok(ast::Expr::List(Spanned::new(
+                        ast::List { items: vec![] },
+                        self.stream.expand_span(span),
+                    )));
+                }
+
+                let mut expr = self.parse_expr()?;
+                if matches!(self.stream.current()?, Some((Token::Comma, _))) {
+                    let mut items = vec![expr];
+                    loop {
+                        if matches!(self.stream.current()?, Some((Token::ParenClose, _))) {
+                            break;
+                        }
+                        expect_token!(self, Token::Comma, "`,`")?;
+                        if matches!(self.stream.current()?, Some((Token::ParenClose, _))) {
+                            break;
+                        }
+                        items.push(self.parse_expr()?);
+                    }
+                    expr = ast::Expr::List(Spanned::new(
+                        ast::List { items },
+                        self.stream.expand_span(span),
+                    ));
+                }
+
                 expect_token!(self, Token::ParenClose, "`)`")?;
                 Ok(expr)
             }
@@ -385,6 +475,10 @@ impl<'a> Parser<'a> {
     }
 
     pub fn parse_expr(&mut self) -> Result<ast::Expr<'a>, Error> {
+        self.parse_ifexpr()
+    }
+
+    pub fn parse_expr_noif(&mut self) -> Result<ast::Expr<'a>, Error> {
         self.parse_or()
     }
 
@@ -423,26 +517,91 @@ impl<'a> Parser<'a> {
         }
     }
 
-    fn parse_assign_target(&mut self) -> Result<&'a str, Error> {
-        let (target, _) = expect_token!(self, Token::Ident(name) => name, "identifier")?;
-        if RESERVED_NAMES.contains(&target) {
-            syntax_error!("cannot assign to reserved variable name {}", target);
+    fn parse_assign_name(&mut self) -> Result<ast::Expr<'a>, Error> {
+        let (id, span) = expect_token!(self, Token::Ident(name) => name, "identifier")?;
+        if RESERVED_NAMES.contains(&id) {
+            syntax_error!("cannot assign to reserved variable name {}", id);
+        }
+        Ok(ast::Expr::Var(ast::Spanned::new(ast::Var { id }, span)))
+    }
+
+    fn parse_assignment(&mut self) -> Result<ast::Expr<'a>, Error> {
+        let span = self.stream.current_span();
+        let mut items = Vec::new();
+        let mut is_tuple = false;
+
+        loop {
+            if !items.is_empty() {
+                expect_token!(self, Token::Comma, "`,`")?;
+            }
+            if matches!(
+                self.stream.current()?,
+                Some((Token::ParenClose, _))
+                    | Some((Token::VariableEnd(..), _))
+                    | Some((Token::BlockEnd(..), _))
+                    | Some((Token::Ident("in"), _))
+            ) {
+                break;
+            }
+            items.push(
+                if matches!(self.stream.current()?, Some((Token::ParenOpen, _))) {
+                    self.stream.next()?;
+                    let rv = self.parse_assignment()?;
+                    expect_token!(self, Token::ParenClose, "`)`")?;
+                    rv
+                } else {
+                    self.parse_assign_name()?
+                },
+            );
+            if matches!(self.stream.current()?, Some((Token::Comma, _))) {
+                is_tuple = true;
+            } else {
+                break;
+            }
+        }
+
+        if !is_tuple && items.len() == 1 {
+            Ok(items.into_iter().next().unwrap())
+        } else {
+            Ok(ast::Expr::List(Spanned::new(
+                ast::List { items },
+                self.stream.expand_span(span),
+            )))
         }
-        Ok(target)
     }
 
     fn parse_for_stmt(&mut self) -> Result<ast::ForLoop<'a>, Error> {
-        let target = self.parse_assign_target()?;
+        let target = self.parse_assignment()?;
         expect_token!(self, Token::Ident("in"), "in")?;
-        let iter = self.parse_expr()?;
+        let iter = self.parse_expr_noif()?;
+        let filter_expr = if matches!(self.stream.current()?, Some((Token::Ident("if"), _))) {
+            self.stream.next()?;
+            Some(self.parse_expr()?)
+        } else {
+            None
+        };
         expect_token!(self, Token::BlockEnd(..), "end of block")?;
-        let body = self.subparse(|tok| matches!(tok, Token::Ident("endfor")))?;
+        let body =
+            self.subparse(|tok| matches!(tok, Token::Ident("endfor") | Token::Ident("else")))?;
+        let else_body = if matches!(self.stream.current()?, Some((Token::Ident("else"), _))) {
+            self.stream.next()?;
+            expect_token!(self, Token::BlockEnd(..), "end of block")?;
+            self.subparse(|tok| matches!(tok, Token::Ident("endfor")))?
+        } else {
+            Vec::new()
+        };
         self.stream.next()?;
-        Ok(ast::ForLoop { target, iter, body })
+        Ok(ast::ForLoop {
+            target,
+            iter,
+            filter_expr,
+            body,
+            else_body,
+        })
     }
 
     fn parse_if_cond(&mut self) -> Result<ast::IfCond<'a>, Error> {
-        let expr = self.parse_expr()?;
+        let expr = self.parse_expr_noif()?;
         expect_token!(self, Token::BlockEnd(..), "end of block")?;
         let true_body = self.subparse(|tok| {
             matches!(
@@ -478,7 +637,10 @@ impl<'a> Parser<'a> {
             if !assignments.is_empty() {
                 expect_token!(self, Token::Comma, "comma")?;
             }
-            let target = self.parse_assign_target()?;
+            let target = match self.parse_assign_name()? {
+                ast::Expr::Var(var) => var.id,
+                _ => panic!("unexpected node from assignment parse"),
+            };
             expect_token!(self, Token::Assign, "assignment operator")?;
             let expr = self.parse_expr()?;
             assignments.push((target, expr));
