@@ -28,14 +28,13 @@
 //! ```
 //!
 //! MiniJinja will perform the necessary conversions automatically via the
-//! [`ValueArgs`](crate::value::ValueArgs) trait.
+//! [`FunctionArgs`](crate::value::FunctionArgs) trait.
 use std::collections::BTreeMap;
-use std::convert::TryFrom;
 use std::sync::Arc;
 
 use crate::environment::Environment;
-use crate::error::{Error, ErrorKind};
-use crate::value::{Value, ValueArgs};
+use crate::error::Error;
+use crate::value::{ArgType, FunctionArgs, Value};
 
 type TestFunc =
     dyn Fn(&Environment, Value, Vec<Value>) -> Result<bool, Error> + Sync + Send + 'static;
@@ -74,19 +73,14 @@ impl BoxedTest {
     pub fn new<F, V, Args>(f: F) -> BoxedTest
     where
         F: Test<V, Args>,
-        V: TryFrom<Value>,
-        Args: ValueArgs,
+        V: ArgType,
+        Args: FunctionArgs,
     {
         BoxedTest(Arc::new(move |env, value, args| -> Result<bool, Error> {
             f.perform(
                 env,
-                TryFrom::try_from(value).map_err(|_| {
-                    Error::new(
-                        ErrorKind::ImpossibleOperation,
-                        "imcompatible value for filter",
-                    )
-                })?,
-                ValueArgs::from_values(args)?,
+                ArgType::from_value(Some(value))?,
+                FunctionArgs::from_values(args)?,
             )
         }))
     }
@@ -102,48 +96,86 @@ impl BoxedTest {
     }
 }
 
-/// Checks if a value is odd.
-pub fn is_odd(_env: &Environment, v: Value) -> Result<bool, Error> {
-    Ok(v.as_primitive()
-        .and_then(|x| x.as_i128())
-        .map_or(false, |x| x % 2 != 0))
-}
-
-/// Checks if a value is even.
-pub fn is_even(_env: &Environment, v: Value) -> Result<bool, Error> {
-    Ok(v.as_primitive()
-        .and_then(|x| x.as_i128())
-        .map_or(false, |x| x % 2 == 0))
-}
-
-/// Checks if a value is undefined.
-pub fn is_undefined(_env: &Environment, v: Value) -> Result<bool, Error> {
-    Ok(v.is_undefined())
-}
-
-/// Checks if a value is defined.
-pub fn is_defined(_env: &Environment, v: Value) -> Result<bool, Error> {
-    Ok(!v.is_undefined())
-}
-
-pub(crate) fn get_default_tests() -> BTreeMap<&'static str, BoxedTest> {
+pub(crate) fn get_builtin_tests() -> BTreeMap<&'static str, BoxedTest> {
+    #[allow(unused_mut)]
     let mut rv = BTreeMap::new();
-    rv.insert("odd", BoxedTest::new(is_odd));
-    rv.insert("even", BoxedTest::new(is_even));
-    rv.insert("undefined", BoxedTest::new(is_undefined));
-    rv.insert("defined", BoxedTest::new(is_defined));
+    #[cfg(feature = "builtin_tests")]
+    {
+        rv.insert("odd", BoxedTest::new(is_odd));
+        rv.insert("even", BoxedTest::new(is_even));
+        rv.insert("undefined", BoxedTest::new(is_undefined));
+        rv.insert("defined", BoxedTest::new(is_defined));
+        rv.insert("number", BoxedTest::new(is_number));
+        rv.insert("string", BoxedTest::new(is_string));
+        rv.insert("sequence", BoxedTest::new(is_sequence));
+        rv.insert("mapping", BoxedTest::new(is_mapping));
+    }
     rv
 }
 
-#[test]
-fn test_basics() {
-    fn test(_: &Environment, a: u32, b: u32) -> Result<bool, Error> {
-        Ok(a == b)
+#[cfg(feature = "builtin_tests")]
+mod builtins {
+    use super::*;
+
+    use crate::value::ValueKind;
+
+    /// Checks if a value is odd.
+    pub fn is_odd(_env: &Environment, v: Value) -> Result<bool, Error> {
+        Ok(v.as_primitive()
+            .and_then(|x| x.as_i128())
+            .map_or(false, |x| x % 2 != 0))
     }
 
-    let env = Environment::new();
-    let bx = BoxedTest::new(test);
-    assert!(bx
-        .perform(&env, Value::from(23), vec![Value::from(23)])
-        .unwrap());
+    /// Checks if a value is even.
+    pub fn is_even(_env: &Environment, v: Value) -> Result<bool, Error> {
+        Ok(v.as_primitive()
+            .and_then(|x| x.as_i128())
+            .map_or(false, |x| x % 2 == 0))
+    }
+
+    /// Checks if a value is undefined.
+    pub fn is_undefined(_env: &Environment, v: Value) -> Result<bool, Error> {
+        Ok(v.is_undefined())
+    }
+
+    /// Checks if a value is defined.
+    pub fn is_defined(_env: &Environment, v: Value) -> Result<bool, Error> {
+        Ok(!v.is_undefined())
+    }
+
+    /// Checks if this value is a number.
+    pub fn is_number(_env: &Environment, v: Value) -> Result<bool, Error> {
+        Ok(matches!(v.kind(), ValueKind::Number))
+    }
+
+    /// Checks if this value is a string.
+    pub fn is_string(_env: &Environment, v: Value) -> Result<bool, Error> {
+        Ok(matches!(v.kind(), ValueKind::String))
+    }
+
+    /// Checks if this value is a sequence
+    pub fn is_sequence(_env: &Environment, v: Value) -> Result<bool, Error> {
+        Ok(matches!(v.kind(), ValueKind::Seq))
+    }
+
+    /// Checks if this value is a mapping
+    pub fn is_mapping(_env: &Environment, v: Value) -> Result<bool, Error> {
+        Ok(matches!(v.kind(), ValueKind::Map))
+    }
+
+    #[test]
+    fn test_basics() {
+        fn test(_: &Environment, a: u32, b: u32) -> Result<bool, Error> {
+            Ok(a == b)
+        }
+
+        let env = Environment::new();
+        let bx = BoxedTest::new(test);
+        assert!(bx
+            .perform(&env, Value::from(23), vec![Value::from(23)])
+            .unwrap());
+    }
 }
+
+#[cfg(feature = "builtin_tests")]
+pub use self::builtins::*;
