@@ -38,9 +38,9 @@
 //!
 //! # Value Function Arguments
 //!
-//! [Filters](crate::filters) and (tests)[crate::tests] can take values as arguments
+//! [Filters](crate::filters) and [tests](crate::tests) can take values as arguments
 //! but optionally also rust types directly.  This conversion for function arguments
-//! is performed by the [`ValueArgs`] trait.
+//! is performed by the [`FunctionArgs`] trait.
 //!
 //! # Memory Management
 //!
@@ -59,12 +59,13 @@
 // this module is based on the content module in insta which in turn is based
 // on the content module in serde::private::ser.
 use std::borrow::Cow;
+use std::cell::RefCell;
 use std::cmp::Ordering;
 use std::collections::BTreeMap;
 use std::convert::TryFrom;
 use std::fmt::{self, Write};
 use std::marker::PhantomData;
-use std::sync::atomic::{self, AtomicBool};
+use std::sync::atomic::{self, AtomicBool, AtomicUsize};
 use std::sync::Arc;
 
 use serde::ser::{self, Serialize, Serializer};
@@ -77,11 +78,12 @@ pub(crate) type RcType<T> = Arc<T>;
 
 // We use in-band signalling to roundtrip some internal values.  This is
 // not ideal but unfortunately there is no better system in serde today.
-const SAFESTRING_MARKER: &str = "\x01__minijinja_SafeString";
-const UNDEFINED_MARKER: &str = "\x01__minijinja_Undefined";
+const VALUE_HANDLE_MARKER: &str = "\x01__minijinja_ValueHandle";
 
 thread_local! {
     static INTERNAL_SERIALIZATION: AtomicBool = AtomicBool::new(false);
+    static LAST_VALUE_HANDLE: AtomicUsize = AtomicUsize::new(0);
+    static VALUE_HANDLES: RefCell<BTreeMap<usize, Value>> = RefCell::new(BTreeMap::new());
 }
 
 fn in_internal_serialization() -> bool {
@@ -92,34 +94,48 @@ fn in_internal_serialization() -> bool {
 ///
 /// Since it's more convenient to write filters and tests with concrete
 /// types instead of values, this helper trait exists to automatically
-/// perform this conversion.
-pub trait ValueArgs: Sized {
+/// perform this conversion.  It is implemented for functions up to an
+/// arity of 5 parameters.
+///
+/// For each argument the conversion is performed via the [`ArgType`]
+/// trait which is implemented for some primitive concrete types as well
+/// as these types wrapped in [`Option`].
+pub trait FunctionArgs: Sized {
     /// Converts to function arguments from a slice of values.
     fn from_values(values: Vec<Value>) -> Result<Self, Error>;
 }
 
+/// A trait implemented by all filter/test argument types.
+///
+/// This trait is the companion to [`FunctionArgs`].  It's passed an
+/// `Option<Value>` where `Some` means the argument was provided or
+/// `None` if it was not.  This is used to implement optional arguments
+/// to functions.
+pub trait ArgType: Sized {
+    fn from_value(value: Option<Value>) -> Result<Self, Error>;
+}
+
 macro_rules! tuple_impls {
     ( $( $name:ident )* ) => {
-        impl<$($name: TryFrom<Value>,)*> ValueArgs for ($($name,)*) {
+        impl<$($name: ArgType,)*> FunctionArgs for ($($name,)*) {
             fn from_values(values: Vec<Value>) -> Result<Self, Error> {
-                #[allow(non_snake_case)]
-                match values.as_slice() {
-                    &[$(ref $name,)*] => Ok((
-                        $(
-                            TryFrom::try_from($name.clone()).map_err(|_| {
-                                Error::new(
-                                    ErrorKind::ImpossibleOperation,
-                                    "incompatible arguemnt type for filter",
-                                )
-                            })?,
-                        )*
-                    )),
-                    _ => {
-                        return Err(Error::new(
-                            ErrorKind::InvalidFilterArguments,
-                            "invalid argument count to filter",
-                        ))
-                    }
+                #![allow(non_snake_case, unused)]
+                let arg_count = 0 $(
+                    + { let $name = (); 1 }
+                )*;
+                if values.len() > arg_count {
+                    return Err(Error::new(
+                        ErrorKind::InvalidArguments,
+                        "received unexpected extra arguments",
+                    ));
+                }
+                {
+                    let mut idx = 0;
+                    $(
+                        let $name = ArgType::from_value(values.get(idx).cloned())?;
+                        idx += 1;
+                    )*
+                    Ok(( $($name,)* ))
                 }
             }
         }
@@ -131,8 +147,9 @@ tuple_impls! { A }
 tuple_impls! { A B }
 tuple_impls! { A B C }
 tuple_impls! { A B C D }
+tuple_impls! { A B C D E }
 
-impl ValueArgs for Vec<Value> {
+impl FunctionArgs for Vec<Value> {
     fn from_values(values: Vec<Value>) -> Result<Self, Error> {
         Ok(values)
     }
@@ -599,6 +616,34 @@ pub(crate) fn string_concat(left: Value, right: &Value) -> Value {
     }
 }
 
+/// Implements a containment operation on values.
+pub(crate) fn contains(container: &Value, value: &Value) -> Result<Value, Error> {
+    if let Value(Repr::Shared(ref cplx)) = container {
+        match **cplx {
+            Shared::Seq(ref values) => return Ok(Value::from(values.contains(value))),
+            Shared::Map(ref map) => {
+                let key = match Key::try_from(value.clone()) {
+                    Ok(key) => key,
+                    Err(_) => return Ok(Value::from(false)),
+                };
+                return Ok(Value::from(map.get(&key).is_some()));
+            }
+            Shared::String(ref s) | Shared::SafeString(ref s) => {
+                return Ok(Value::from(if let Some(s2) = value.as_str() {
+                    s.contains(&s2)
+                } else {
+                    s.contains(&value.to_string())
+                }));
+            }
+            _ => {}
+        }
+    }
+    Err(Error::new(
+        ErrorKind::ImpossibleOperation,
+        "cannot perform a containment check on this value",
+    ))
+}
+
 macro_rules! primitive_try_from {
     ($ty:ident, {
         $($pat:pat => $expr:expr,)*
@@ -617,6 +662,30 @@ macro_rules! primitive_try_from {
                 })
             }
         }
+
+        impl ArgType for $ty {
+            fn from_value(value: Option<Value>) -> Result<Self, Error> {
+                match value {
+                    Some(value) => TryFrom::try_from(value.clone()),
+                    None => Err(Error::new(ErrorKind::UndefinedError, concat!("missing argument")))
+                }
+            }
+        }
+
+        impl ArgType for Option<$ty> {
+            fn from_value(value: Option<Value>) -> Result<Self, Error> {
+                match value {
+                    Some(value) => {
+                        if value.is_undefined() || value.is_none() {
+                            Ok(None)
+                        } else {
+                            TryFrom::try_from(value.clone()).map(Some)
+                        }
+                    }
+                    None => Ok(None),
+                }
+            }
+        }
     }
 }
 
@@ -639,6 +708,7 @@ primitive_int_try_from!(i16);
 primitive_int_try_from!(i32);
 primitive_int_try_from!(i64);
 primitive_int_try_from!(i128);
+primitive_int_try_from!(usize);
 
 primitive_try_from!(bool, {
     Primitive::Bool(val) => val,
@@ -648,6 +718,40 @@ primitive_try_from!(f64, {
     Primitive::F64(val) => val,
 });
 
+macro_rules! infallible_conversion {
+    ($ty:ty) => {
+        impl ArgType for $ty {
+            fn from_value(value: Option<Value>) -> Result<Self, Error> {
+                match value {
+                    Some(value) => Ok(value.clone().into()),
+                    None => Err(Error::new(
+                        ErrorKind::UndefinedError,
+                        concat!("missing argument"),
+                    )),
+                }
+            }
+        }
+
+        impl ArgType for Option<$ty> {
+            fn from_value(value: Option<Value>) -> Result<Self, Error> {
+                match value {
+                    Some(value) => {
+                        if value.is_undefined() || value.is_none() {
+                            Ok(None)
+                        } else {
+                            Ok(Some(value.clone().into()))
+                        }
+                    }
+                    None => Ok(None),
+                }
+            }
+        }
+    };
+}
+
+infallible_conversion!(String);
+infallible_conversion!(Value);
+
 impl From<Value> for String {
     fn from(val: Value) -> Self {
         val.to_string()
@@ -761,6 +865,11 @@ impl Value {
         matches!(&self.0, Repr::Undefined)
     }
 
+    /// Returns `true` if this value is none.
+    pub fn is_none(&self) -> bool {
+        matches!(&self.0, Repr::None)
+    }
+
     /// Returns the length of the contained value.
     #[allow(clippy::len_without_is_empty)]
     pub fn len(&self) -> Option<usize> {
@@ -857,18 +966,6 @@ impl Value {
         ))
     }
 
-    pub(crate) fn into_string(self) -> String {
-        if let Repr::Shared(arc) = self.0 {
-            match RcType::try_unwrap(arc) {
-                Ok(Shared::String(s)) | Ok(Shared::SafeString(s)) => s,
-                Ok(other) => other.to_string(),
-                Err(arc) => arc.to_string(),
-            }
-        } else {
-            self.to_string()
-        }
-    }
-
     pub(crate) fn try_into_vec(self) -> Result<Vec<Value>, Error> {
         if let Repr::Shared(arc) = self.0 {
             match RcType::try_unwrap(arc) {
@@ -887,6 +984,33 @@ impl Value {
         ))
     }
 
+    #[cfg(feature = "builtin_filters")]
+    pub(crate) fn try_into_pairs(self) -> Result<Vec<Value>, Error> {
+        if let Repr::Shared(arc) = self.0 {
+            match RcType::try_unwrap(arc) {
+                Ok(Shared::Map(v)) => {
+                    return Ok(v
+                        .into_iter()
+                        .map(|(k, v)| Value::from(vec![Value::from(k), v]))
+                        .collect())
+                }
+                Ok(_) => {}
+                Err(arc) => {
+                    if let Shared::Map(v) = &*arc {
+                        return Ok(v
+                            .iter()
+                            .map(|(k, v)| Value::from(vec![Value::from(k.clone()), v.clone()]))
+                            .collect());
+                    }
+                }
+            }
+        }
+        Err(Error::new(
+            ErrorKind::ImpossibleOperation,
+            "cannot convert value into pair list",
+        ))
+    }
+
     /// Iterates over the value.
     pub(crate) fn iter(&self) -> ValueIterator {
         let value = self.clone();
@@ -917,6 +1041,16 @@ impl Serialize for Value {
     where
         S: Serializer,
     {
+        // enable round tripping of values
+        if in_internal_serialization() {
+            use serde::ser::SerializeStruct;
+            let handle = LAST_VALUE_HANDLE.with(|x| x.fetch_add(1, atomic::Ordering::Relaxed));
+            VALUE_HANDLES.with(|handles| handles.borrow_mut().insert(handle, self.clone()));
+            let mut s = serializer.serialize_struct(VALUE_HANDLE_MARKER, 1)?;
+            s.serialize_field("handle", &handle)?;
+            return s.end();
+        }
+
         match self.0 {
             Repr::Bool(b) => serializer.serialize_bool(b),
             Repr::U64(u) => serializer.serialize_u64(u),
@@ -924,28 +1058,12 @@ impl Serialize for Value {
             Repr::F64(f) => serializer.serialize_f64(f),
             Repr::Char(c) => serializer.serialize_char(c),
             Repr::None => serializer.serialize_unit(),
-            Repr::Undefined => {
-                use serde::ser::SerializeStruct;
-                if in_internal_serialization() {
-                    serializer.serialize_struct(UNDEFINED_MARKER, 1)?.end()
-                } else {
-                    serializer.serialize_unit()
-                }
-            }
+            Repr::Undefined => serializer.serialize_unit(),
             Repr::Shared(ref cplx) => match **cplx {
                 Shared::U128(u) => serializer.serialize_u128(u),
                 Shared::I128(i) => serializer.serialize_i128(i),
                 Shared::String(ref s) => serializer.serialize_str(s),
-                Shared::SafeString(ref val) => {
-                    use serde::ser::SerializeStruct;
-                    if in_internal_serialization() {
-                        let mut s = serializer.serialize_struct(SAFESTRING_MARKER, 1)?;
-                        s.serialize_field("value", val)?;
-                        s.end()
-                    } else {
-                        serializer.serialize_str(val)
-                    }
-                }
+                Shared::SafeString(ref val) => serializer.serialize_str(val),
                 Shared::Bytes(ref b) => serializer.serialize_bytes(b),
                 Shared::Seq(ref elements) => elements.serialize(serializer),
                 Shared::Map(ref entries) => {
@@ -1369,10 +1487,18 @@ where
 
     fn end(self) -> Result<Value, E> {
         match self.name {
-            SAFESTRING_MARKER if self.fields.len() == 1 => Ok(Value::from_safe_string(
-                self.fields.into_iter().next().unwrap().1.into_string(),
-            )),
-            UNDEFINED_MARKER => Ok(Repr::Undefined.into()),
+            VALUE_HANDLE_MARKER => {
+                let handle_id = match self.fields.get("handle") {
+                    Some(&Value(Repr::U64(handle_id))) => handle_id as usize,
+                    _ => panic!("bad handle reference in value roundtrip"),
+                };
+                Ok(VALUE_HANDLES.with(|handles| {
+                    let mut handles = handles.borrow_mut();
+                    handles
+                        .remove(&handle_id)
+                        .expect("value handle not in registry")
+                }))
+            }
             _ => Ok(Shared::Struct(self.fields).into()),
         }
     }
@@ -1530,6 +1656,40 @@ fn test_undefined_roundtrip() {
     assert!(v2.is_undefined());
 }
 
+#[test]
+fn test_dynamic_object_roundtrip() {
+    #[derive(Debug)]
+    struct X(AtomicUsize);
+
+    impl fmt::Display for X {
+        fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+            write!(f, "{}", self.0.load(atomic::Ordering::Relaxed))
+        }
+    }
+
+    impl DynamicObject for X {
+        fn get_attr(&self, name: &str) -> Option<Value> {
+            match name {
+                "value" => Some(Value::from(self.0.load(atomic::Ordering::Relaxed))),
+                _ => None,
+            }
+        }
+
+        fn fields(&self) -> &'static [&'static str] {
+            &["value"]
+        }
+    }
+
+    let x = RcType::new(X(Default::default()));
+    let x_value = Value::from_dynamic(x.clone());
+    x.0.fetch_add(42, atomic::Ordering::Relaxed);
+    let x_clone = Value::from_serializable(&x_value);
+    x.0.fetch_add(23, atomic::Ordering::Relaxed);
+
+    assert_eq!(x_value.to_string(), "65");
+    assert_eq!(x_clone.to_string(), "65");
+}
+
 #[test]
 fn test_string_key_lookup() {
     let mut m = BTreeMap::new();
