@@ -70,8 +70,8 @@ impl fmt::Display for LoopState {
 
 #[derive(Debug)]
 pub struct Loop<'source> {
-    target_name: &'source str,
-    current_value: Value,
+    locals: BTreeMap<&'source str, Value>,
+    with_loop_var: bool,
     iterator: ValueIterator,
     controller: RcType<LoopState>,
 }
@@ -123,6 +123,11 @@ pub struct Context<'source, 'context> {
 }
 
 impl<'source, 'context> Context<'source, 'context> {
+    /// Stores a variable in the context.
+    pub fn store(&mut self, key: &'source str, value: Value) {
+        self.current_loop().locals.insert(key, value);
+    }
+
     /// Looks up a variable in the context.
     pub fn lookup(&self, key: &str) -> Option<Value> {
         for ctx in self.stack.iter().rev() {
@@ -133,15 +138,15 @@ impl<'source, 'context> Context<'source, 'context> {
                 Frame::Isolate { value } => (value, false),
                 Frame::Merge { value } => (value, true),
                 Frame::Loop(Loop {
-                    target_name,
-                    current_value,
+                    locals,
                     controller,
+                    with_loop_var,
                     ..
                 }) => {
-                    if key == *target_name {
-                        return Some(current_value.clone());
-                    } else if key == "loop" {
+                    if *with_loop_var && key == "loop" {
                         return Some(Value::from_dynamic(controller.clone()));
+                    } else if let Some(value) = locals.get(key) {
+                        return Some(value.clone());
                     }
                     continue;
                 }
@@ -243,33 +248,38 @@ impl<'env, 'source> Vm<'env, 'source> {
         let mut auto_escape = initial_auto_escape;
         let mut auto_escape_stack = vec![];
 
+        macro_rules! bail {
+            ($err:expr) => {{
+                let mut err = $err;
+                if let Some((filename, lineno)) = instructions.get_location(pc) {
+                    err.set_location(filename, lineno);
+                }
+                return Err(err);
+            }};
+        }
+
         macro_rules! try_ctx {
             ($expr:expr) => {
                 match $expr {
                     Ok(rv) => rv,
-                    Err(mut err) => {
-                        if let Some((filename, lineno)) = instructions.get_location(pc) {
-                            err.set_location(filename, lineno);
-                        }
-                        return Err(err);
-                    }
+                    Err(err) => bail!(err),
                 }
             };
         }
 
         macro_rules! func_binop {
             ($method:ident) => {{
-                let a = stack.pop();
                 let b = stack.pop();
-                stack.push(try_ctx!(value::$method(&b, &a)));
+                let a = stack.pop();
+                stack.push(try_ctx!(value::$method(&a, &b)));
             }};
         }
 
         macro_rules! op_binop {
             ($op:tt) => {{
-                let a = stack.pop();
                 let b = stack.pop();
-                stack.push(Value::from(b $op a));
+                let a = stack.pop();
+                stack.push(Value::from(a $op b));
             }};
         }
 
@@ -300,6 +310,9 @@ impl<'env, 'source> Vm<'env, 'source> {
                 Instruction::Emit => {
                     try_ctx!(self.env.finalize(&stack.pop(), auto_escape, output));
                 }
+                Instruction::StoreLocal(name) => {
+                    context.store(name, stack.pop());
+                }
                 Instruction::Lookup(name) => {
                     stack.push(context.lookup(name).unwrap_or(Value::UNDEFINED));
                 }
@@ -332,6 +345,24 @@ impl<'env, 'source> Vm<'env, 'source> {
                     v.reverse();
                     stack.push(v.into());
                 }
+                Instruction::UnpackList(count) => {
+                    let mut v = try_ctx!(stack.pop().try_into_vec());
+                    if v.len() != *count {
+                        bail!(Error::new(
+                            ErrorKind::ImpossibleOperation,
+                            "sequence of wrong length"
+                        ));
+                    }
+                    for _ in 0..*count {
+                        stack.push(v.pop().unwrap());
+                    }
+                }
+                Instruction::ListAppend => {
+                    let item = stack.pop();
+                    let mut list = try_ctx!(stack.pop().try_into_vec());
+                    list.push(item);
+                    stack.push(Value::from(list));
+                }
                 Instruction::Add => func_binop!(add),
                 Instruction::Sub => func_binop!(sub),
                 Instruction::Mul => func_binop!(mul),
@@ -353,6 +384,11 @@ impl<'env, 'source> Vm<'env, 'source> {
                     let b = stack.pop();
                     stack.push(value::string_concat(b, &a));
                 }
+                Instruction::In => {
+                    let container = stack.pop();
+                    let value = stack.pop();
+                    stack.push(try_ctx!(value::contains(&container, &value)));
+                }
                 Instruction::Neg => {
                     let a = stack.pop();
                     stack.push(try_ctx!(value::neg(&a)));
@@ -364,14 +400,14 @@ impl<'env, 'source> Vm<'env, 'source> {
                 Instruction::PopFrame => {
                     context.pop_frame();
                 }
-                Instruction::PushLoop(target_name) => {
+                Instruction::PushLoop(with_loop_var) => {
                     let iterable = stack.pop();
                     let iterator = iterable.iter();
                     let len = iterator.len();
                     context.push_frame(Frame::Loop(Loop {
-                        target_name,
-                        current_value: Value::UNDEFINED,
+                        locals: BTreeMap::new(),
                         iterator,
+                        with_loop_var: *with_loop_var,
                         controller: RcType::new(LoopState {
                             idx: AtomicUsize::new(!0usize),
                             len: AtomicUsize::new(len),
@@ -381,8 +417,10 @@ impl<'env, 'source> Vm<'env, 'source> {
                 Instruction::Iterate(jump_target) => {
                     let l = context.current_loop();
                     l.controller.idx.fetch_add(1, Ordering::Relaxed);
-                    l.current_value = match l.iterator.next() {
-                        Some(item) => item,
+                    match l.iterator.next() {
+                        Some(item) => {
+                            stack.push(item);
+                        }
                         None => {
                             pc = *jump_target;
                             continue;
@@ -403,6 +441,7 @@ impl<'env, 'source> Vm<'env, 'source> {
                 Instruction::JumpIfFalseOrPop(jump_target) => {
                     if !stack.peek().is_true() {
                         pc = *jump_target;
+                        continue;
                     } else {
                         stack.pop();
                     }
@@ -483,7 +522,7 @@ impl<'env, 'source> Vm<'env, 'source> {
                             }
                         }
                         _ => {
-                            return Err(Error::new(
+                            bail!(Error::new(
                                 ErrorKind::ImpossibleOperation,
                                 "invalid value to autoescape tag",
                             ));
@@ -520,7 +559,7 @@ impl<'env, 'source> Vm<'env, 'source> {
                             panic!("attempted to super unreferenced block");
                         }
                     } else {
-                        return Err(Error::new(
+                        bail!(Error::new(
                             ErrorKind::ImpossibleOperation,
                             format!("unknown function {}", function_name),
                         ));
@@ -536,11 +575,17 @@ impl<'env, 'source> Vm<'env, 'source> {
                     let _obj = stack.pop();
                     // TODO: this is something that doesn't make too much sense in the
                     // engine today.
-                    return Err(Error::new(
+                    bail!(Error::new(
                         ErrorKind::ImpossibleOperation,
                         "objects cannot be called directly",
                     ));
                 }
+                Instruction::DupTop => {
+                    stack.push(stack.peek().clone());
+                }
+                Instruction::DiscardTop => {
+                    stack.pop();
+                }
                 Instruction::Nop => {}
             }
             pc += 1;
