@@ -6,10 +6,9 @@ use minijinja::value::Value;
 fn test_for_loop() {
     let mut c = Compiler::new();
     c.add(Instruction::Lookup("items"));
-    c.start_for_loop("x");
-    c.add(Instruction::Lookup("x"));
+    c.start_for_loop(true);
     c.add(Instruction::Emit);
-    c.end_for_loop();
+    c.end_for_loop(false);
     c.add(Instruction::EmitRaw("!"));
 
     insta::assert_debug_snapshot!(&c);
